#############################################
#############################################
##    Example use of foreach and doParallel
##    Amy Mulick, LSHTM
##    R User Group meeting, 28 June 2019
#############################################
#############################################

rm(list=ls())

require("doParallel")
require("R2jags")



#####################
#foreach
#####################

#simple non-parallel: %do%
a <- foreach(i=1:3) %do% exp(i)
a

b <- foreach(i=1:3, .combine='c') %do% exp(i)
b

c <- foreach(i=1:3, .combine='rbind') %do% exp(i)
c

#####################
#example to parallelise: Bayesian Hierarchical Model
#####################

#data and params for JAGS model
model.file <- system.file(package="R2jags", "model", "schools.txt")
J <- 8.0
y <- c(28.4,7.9,-2.8,6.8,-0.6,0.6,18.0,12.2)
sd <- c(14.9,10.2,16.3,11.0,9.4,11.4,10.4,17.6)

jags.params <- c("mu","sigma","theta")
jags.inits <- function(){
  list("mu"=rnorm(1),"sigma"=runif(1),"theta"=rnorm(J))
}  

  
#reference point: repeat the same analysis 7 times in serial
Start <- Sys.time()

x <- foreach(i=1:7) %do% jags(data=list("y","sd","J"), inits=jags.inits, jags.params,
                              n.iter=100000*i, model.file=model.file)

End <- Sys.time()

print(End - Start)
View(x)



#now parallelise using %dopar% 

#wrong way #1: failure to make a cluster
z <- foreach(i=1:7) %dopar% 
  jags(data=list("y","sd","J"), inits=jags.inits, jags.params,n.iter=100000*i, model.file=model.file)
View(z)


  #Make a cluster using all available 'workers' except 1
  nc <- detectCores()
  nc
  
  cl <- makeCluster(nc-1)
  registerDoParallel(cl)

  
#wrong way #2: failure to send required packages to cluster workers
z <- foreach(i=1:7) %dopar% 
  jags(data=list("y","sd","J"), inits=jags.inits, jags.params,n.iter=100000*i, model.file=model.file)



  
#wrong way #3: failure to send environment variables to cluster workers
z <- foreach(i=1:7, .packages = c("R2jags")) %dopar% 
  jags(data=list("y","sd","J"), inits=jags.inits, jags.params,n.iter=100000*i, model.file=model.file)



#Right way!
Start <- Sys.time()
z <- foreach(i=1:7, .packages = c("R2jags"), .export = c("y","sd","J","jags.inits","jags.params","model.file")) %dopar% 
  jags(data=list("y","sd","J"), inits=jags.inits, jags.params,n.iter=100000*i, model.file=model.file)
End <- Sys.time()
print(End - Start)

View(z)



#Example of the utility of foreach + doParallel (failure to model check not recommended)
zz <- foreach(i=1:7, .packages = c("R2jags"), .export = c("y","sd","J","jags.inits","jags.params","model.file"), .combine='rbind') %dopar% 
  jags(data=list("y","sd","J"), inits=jags.inits, jags.params,n.iter=100000*i, model.file=model.file)$BUGSoutput$median$mu

View(zz)



#stop the cluster
stopCluster(cl)
